//
// Copyright (c) 2007-2008, by YaPB Development Team. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Version: $Id:$
//

#include <core.h>

BotManager::BotManager (void)
{
   // this is a bot manager class constructor

   m_lastWinner = -1;

   m_economicsGood[Team_Terrorist] = true;
   m_economicsGood[Team_Counter] = true;

   memset (m_bots, 0, sizeof (m_bots));
   InitQuota ();
}

BotManager::~BotManager (void)
{
   // this is a bot manager class destructor, do not use GetMaxClients () here !!

   for (int i = 0; i < 32; i++)
   {
      if (m_bots[i])
      {
         delete m_bots[i];
         m_bots[i] = NULL;
      }
   }
}

void BotManager::CallGameEntity (entvars_t *vars)
{
   // this function calls gamedll player() function, in case to create player entity in game

   if (g_isMetamod)
   {
      CALL_GAME_ENTITY (PLID, "player", vars);
      return;
   }

   static EntityPtr_t playerFunction = NULL;

   if (playerFunction == NULL)
      playerFunction = (EntityPtr_t) g_gameLib->GetFunctionAddr ("player");

   if (playerFunction != NULL)
      (*playerFunction) (vars);
}

int BotManager::CreateBot (String name, int skill, int personality, int team, int member)
{
   // this function completely prepares bot entity (edict) for creation, creates team, skill, sets name etc, and
   // then sends result to bot constructor

   edict_t *bot = NULL;
   char outputName[33];

   if (g_numWaypoints < 1) // don't allow creating bots with no waypoints loaded
   {
      CenterPrint ("Map not waypointed. Can't Create Bot");
      return 0;
   }
   else if (g_waypointsChanged) // don't allow creating bots with changed waypoints (distance tables are messed up)
   {
      CenterPrint ("Waypoints has been changed. Load waypoints again...");
      return 0;
   }

   if (skill < 0 || skill > 100)
      skill = g_randGen.Long (g_botVar[Variable_MinSkill]->GetInt (), g_botVar[Variable_MaxSkill]->GetInt ());

   if (skill > 100 || skill < 0)
      skill = g_randGen.Long (0, 100);

   if (personality < 0 || personality > 2)
   {
      int randomPrecent = g_randGen.Long (0, 100);

      if (randomPrecent < 50)
         personality = Personality_Normal;
      else
      {
         if (g_randGen.Long (0, randomPrecent) < randomPrecent * 0.5)
            personality = Personality_Careful;
         else
            personality = Personality_Rusher;
      }
   }

   // setup name
   if (name.IsEmpty ())
   {
      if (!g_botNames.IsEmpty ())
      {
         bool nameFound = false;

         for (int i = 0; i < 8; i++)
         {
            if (nameFound)
               break;

            BotName_t *name = &g_botNames.GetRandomElement ();

            if (!name || name->isUsed)
               continue;

            name->isUsed = nameFound = true;
            strcpy (outputName, name->name);
         }
      }
      else
         sprintf (outputName, "bot%i", g_randGen.Long (0, 100)); // just pick ugly random name
   }
   else
      strncpy (outputName, name, 21);

   if (!IsNullString (g_botVar[Variable_Prefix]->GetString ()) || g_botVar[Variable_SkillTags]->GetBool ())
   {
      char prefixedName[33]; // temp buffer for storing modified name

      if (!IsNullString (g_botVar[Variable_Prefix]->GetString ()))
         sprintf (prefixedName, "%s %s", g_botVar[Variable_Prefix]->GetString (), outputName);

      else if (g_botVar[Variable_SkillTags]->GetBool ())
         sprintf  (prefixedName, "%s (%d)", outputName, skill);

      else if (!IsNullString (g_botVar[Variable_Prefix]->GetString ()) && g_botVar[Variable_SkillTags]->GetBool ())
         sprintf  (prefixedName, "%s %s (%d)", g_botVar[Variable_Prefix]->GetString (), outputName, skill);

      // buffer has been modified, copy to real name
      if (!IsNullString (prefixedName))
         sprintf (outputName, prefixedName);
   }

   if (FNullEnt ((bot = (*g_engfuncs.pfnCreateFakeClient) (outputName))))
   {
      CenterPrint ("Maximum players reached (%d/%d). Unable to create Bot.", GetMaxClients (), GetMaxClients ());
      return 2;
   }

   int index = ENTINDEX (bot) - 1;

   InternalAssert (index >= 0 && index <= 32); // check index
   InternalAssert (m_bots[index] == NULL); // check bot slot

   m_bots[index] = new Bot (bot, skill, personality, team, member);

   if (m_bots == NULL)
      TerminateOnMalloc ();

   if (g_botVar[Variable_Developer]->GetBool ())
      ServerPrint ("Connecting '%s'... (Skill %d)", STRING (bot->v.netname), skill);
   else
      ServerPrint ("Connecting YaPB... (Skill %d)", skill);

   return 1;
}

int BotManager::GetIndex (edict_t *ent)
{
   // this function returns index of bot (using own bot array)

   if (FNullEnt (ent))
      return -1;

   int index = ENTINDEX (ent) - 1;

   if (index < 0 || index >= 32)
      return -1;

   if (m_bots[index] != NULL)
      return index;

   return -1; // if no edict, return -1;
}

Bot *BotManager::GetBot (int index)
{
   // this function finds a bot specified by index, and then returns pointer to it (using own bot array)

   if (index < 0 || index >= 32)
      return NULL;

   if (m_bots[index] != NULL)
      return m_bots[index];

   return NULL; // no bot
}

Bot *BotManager::GetBot (edict_t *ent)
{
   // same as above, but using bot entity

   return GetBot (GetIndex (ent));
}

Bot *BotManager::FindOneValidAliveBot (void)
{
   // this function finds one bot, alive bot :)

   Array <int> foundBots;

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL && IsAlive (m_bots[i]->GetEntity ()))
         foundBots.Push (i);
   }

   if (!foundBots.IsEmpty ())
      return m_bots[foundBots.GetRandomElement ()];

   return NULL;
}

void BotManager::Think (void)
{
   // this function calls think () function for all available at call moment bots, and
   // try to catch internal error if such shit occurs

   int thinkFps = g_botVar[Variable_ThinkFPS]->GetInt ();

   // check ranges
   if (thinkFps <= 0)
   {
      AddLogEntry (true, Log_Error, "Value of yb_thinkfps should be greater than zero.");
      g_botVar[Variable_ThinkFPS]->SetInt (30);
   }

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL)
      {
         if (m_bots[i]->m_thinkTimer < GetWorldTime ())
         {
            // use these try-catch blocks to prevent server crashes when error occurs
#if !defined (NDEBUG) && !defined (_DEBUG)
            try
            {
               m_bots[i]->Think ();
            }
            catch (...)
            {
               // error occurred. kick off all bots and then print a warning message
               RemoveAll ();

               ServerPrintNoTag ("**** INTERNAL BOT ERROR! PLEASE SHUTDOWN AND RESTART YOUR SERVER! ****");
            }
#else
            m_bots[i]->Think ();
#endif
            m_bots[i]->m_thinkTimer = GetWorldTime () + (1 / g_botVar[Variable_ThinkFPS]->GetInt ()) * g_randGen.Float (0.97, 1.05);
         }
      }
   }
}

void BotManager::AddBot (String name, int skill, int personality, int team, int member)
{
   // this function putting bot creation process to queue to prevent engine crashes

   CreateQueue_t queueID;

   // fill the holder
   queueID.name = name;
   queueID.skill = skill;
   queueID.personality = personality;
   queueID.team = team;
   queueID.member = member;

   // put to queue
   m_creationTab.Push (queueID);

   // keep quota number up to date
   if (GetBotsNum () + 1 > g_botVar[Variable_Quota]->GetInt ())
      g_botVar[Variable_Quota]->SetInt (GetBotsNum () + 1);
}

void BotManager::AddBot (String name, String skill, String personality, String team, String member)
{
   // this function is same as the function above, but accept as parameters string instead of integers

   CreateQueue_t queueID;
   const String &any = "*";

   queueID.name = (name.IsEmpty () || (name == any)) ?  String ("\0") : name;
   queueID.skill = (skill.IsEmpty () || (skill == any)) ? -1 : skill.ToInt ();
   queueID.team = (team.IsEmpty () || (team == any)) ? -1 : team.ToInt ();
   queueID.member = (member.IsEmpty () || (member == any)) ? -1 : member.ToInt ();
   queueID.personality = (personality.IsEmpty () || (personality == any)) ? -1 : personality.ToInt ();

   m_creationTab.Push (queueID);

   // keep quota number up to date
   if (GetBotsNum () + 1 > g_botVar[Variable_Quota]->GetInt ())
      g_botVar[Variable_Quota]->SetInt (GetBotsNum () + 1);
}

void BotManager::CheckAutoVacate (edict_t *ent)
{
   // this function sets timer to kick one bot off.

   if (g_botVar[Variable_AutoVacate]->GetBool ())
      RemoveRandom ();
}

void BotManager::MaintainBotQuota (void)
{
   // this function keeps number of bots up to date, and don't allow to maintain bot creation
   // while creation process in process.

   if (!m_creationTab.IsEmpty () && m_maintainTime < GetWorldTime ())
   {
      CreateQueue_t last = m_creationTab.Pop ();
      int resultOfCall = CreateBot (last.name, last.skill, last.personality, last.team, last.member);

      // check the result of creation
      if (resultOfCall == 0)
      {
         m_creationTab.RemoveAll (); // something worng with waypoints, reset tab of creation
         g_botVar[Variable_Quota]->SetInt (0); // reset quota
      }
      else if (resultOfCall == 2)
      {
         m_creationTab.RemoveAll (); // maximum players reached, so set quota to maximum players
         g_botVar[Variable_Quota]->SetInt (GetBotsNum ());
      }
      m_maintainTime = GetWorldTime () + 0.2;
   }

   // now keep bot number up to date
   if (m_maintainTime < GetWorldTime ())
   {
      int botNumber = GetBotsNum ();
      int humanNumber = GetHumansNum ();

      if (botNumber > g_botVar[Variable_Quota]->GetInt ())
         RemoveRandom ();

      if (g_botVar[Variable_AutoVacate]->GetBool ())
      {
         if (botNumber < g_botVar[Variable_Quota]->GetInt () && botNumber < GetMaxClients () - 1)
            AddRandom ();

         if (humanNumber >= GetMaxClients ())
            RemoveRandom ();
      }
      else
      {
         if (botNumber < g_botVar[Variable_Quota]->GetInt () && botNumber < GetMaxClients ())
            AddRandom ();
      }

      int botQuota = g_botVar[Variable_AutoVacate]->GetBool () ? (GetMaxClients () - 1 - (humanNumber + 1)) : GetMaxClients ();

      // check valid range of quota
      if (g_botVar[Variable_Quota]->GetInt () > botQuota)
         g_botVar[Variable_Quota]->SetInt (botQuota);

      else if (g_botVar[Variable_Quota]->GetInt () < 0)
         g_botVar[Variable_Quota]->SetInt (0);

      m_maintainTime = GetWorldTime () + 0.25;
   }
}

void BotManager::InitQuota (void)
{
   m_maintainTime = GetWorldTime () + 2.0;
   m_creationTab.RemoveAll ();
}

void BotManager::FillServer (int selection, int personality, int skill, int numToAdd)
{
   // this function fill server with bots, with specified team & personality

   if (GetBotsNum () >= GetMaxClients () - GetHumansNum ())
      return;

   if (selection == 1 || selection == 2)
   {
      CVAR_SET_STRING ("mp_limitteams", "0");
      CVAR_SET_STRING ("mp_autoteambalance", "0");
   }
   else
      selection = 5;

   char teamDesc[6][12] =
   {
      "",
      {"Terrorists"},
      {"CTs"},
      "",
      "",
      {"Random"},
   };

   int toAdd = numToAdd == -1 ? GetMaxClients () - (GetHumansNum () + GetBotsNum ()) : numToAdd;

   for (int i = 0; i <= toAdd; i++)
   {
      // since we got constant skill from menu (since creation process call automatic), we need to manually
      // randomize skill here, on given skill there.
      int randomizedSkill = 0;

      if (skill >= 0 && skill <= 20)
         randomizedSkill = g_randGen.Long (0, 20);
      else if (skill >= 20 && skill <= 40)
         randomizedSkill = g_randGen.Long (20, 40);
      else if (skill >= 40 && skill <= 60)
         randomizedSkill = g_randGen.Long (40, 60);
      else if (skill >= 60 && skill <= 80)
         randomizedSkill = g_randGen.Long (60, 80);
      else if (skill >= 80 && skill <= 99)
         randomizedSkill = g_randGen.Long (80, 99);
      else if (skill == 100)
         randomizedSkill = skill;

      AddBot ("", randomizedSkill, personality, selection, -1);
   }

   g_botVar[Variable_Quota]->SetInt (toAdd);
   CenterPrint ("Fill Server with %s bots...", &teamDesc[selection][0]);
}

void BotManager::RemoveAll (void)
{
   // this function drops all bot clients from server (this function removes only yapb's)`q

   CenterPrint ("Bots are removed from server.");

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL)  // is this slot used?
         m_bots[i]->Kick ();
   }
   m_creationTab.RemoveAll ();

   // reset cvars
   g_botVar[Variable_Quota]->SetInt (0);
   g_botVar[Variable_AutoVacate]->SetInt (0);
}

void BotManager::RemoveFromTeam (Team_t team, bool removeAll)
{
   // this function remove random bot from specified team (if removeAll value = 1 then removes all players from team)

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL && team == GetTeam (m_bots[i]->GetEntity ()))
      {
         m_bots[i]->Kick ();

         if (!removeAll)
            break;
      }
   }
}

void BotManager::RemoveMenu (edict_t *ent, int selection)
{
   // this function displays remove bot menu to specified entity (this function show's only yapb's).

   if (selection > 4 || selection < 1)
      return;

   char tempBuffer[1024], buffer[1024];

   memset (tempBuffer, 0, sizeof (tempBuffer));
   memset (buffer, 0, sizeof (buffer));

   int validSlots = (selection == 4) ? (1 << 9) : ((1 << 8) | (1 << 9));

   for (int i = ((selection - 1) * 8); i < selection * 8; i++)
   {
      if ((m_bots[i] != NULL) && !FNullEnt (m_bots[i]->GetEntity ()))
      {
         validSlots |= 1 << (i - ((selection - 1) * 8));
         sprintf (buffer, "%s %1.1d. %s%s\n", buffer, i - ((selection - 1) * 8) + 1, STRING (m_bots[i]->pev->netname), GetTeam (m_bots[i]->GetEntity ()) == Team_Counter ? " \\y(CT)\\w" : " \\r(T)\\w");
      }
      else
         sprintf (buffer, "%s\\d %1.1d. Not a YaPB\\w\n", buffer, i - ((selection - 1) * 8) + 1);
   }

   sprintf (tempBuffer, "\\yYaPB Remove Menu (%d/4):\\w\n\n%s\n%s 0. Back", selection, buffer, (selection == 4) ? "" : " 9. More...\n");

   switch (selection)
   {
   case 1:
      g_menus[14].validSlots = validSlots & static_cast <unsigned int> (-1);
      g_menus[14].menuText = tempBuffer;

      DisplayMenuToClient (ent, &g_menus[14]);
      break;

   case 2:
      g_menus[15].validSlots = validSlots & static_cast <unsigned int> (-1);
      g_menus[15].menuText = tempBuffer;

      DisplayMenuToClient (ent, &g_menus[15]);
      break;

    case 3:
      g_menus[16].validSlots = validSlots & static_cast <unsigned int> (-1);
      g_menus[16].menuText = tempBuffer;

      DisplayMenuToClient (ent, &g_menus[16]);
      break;

   case 4:
      g_menus[17].validSlots = validSlots & static_cast <unsigned int> (-1);
      g_menus[17].menuText = tempBuffer;

      DisplayMenuToClient (ent, &g_menus[17]);
      break;
   }
}

void BotManager::KillAll (int team)
{
   // this function kills all bots on server (only this dll controlled bots)

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL)
      {
         if (team != -1 && team != GetTeam (m_bots[i]->GetEntity ()))
            continue;

         m_bots[i]->Kill ();
      }
   }
   CenterPrint ("All bots are killed.");
}

void BotManager::RemoveRandom (void)
{
   // this function removes random bot from server (only yapb's)

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL)  // is this slot used?
      {
         m_bots[i]->Kick ();
         break;
      }
   }
}

void BotManager::SetWeaponMode (int selection)
{
   // this function sets bots weapon mode

   int tabMapStandart[7][Const_NumWeapons] =
   {
      {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}, // Knife only
      {-1,-1,-1, 2, 2, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}, // Pistols only
      {-1,-1,-1,-1,-1,-1,-1, 2, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}, // Shotgun only
      {-1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 1, 2, 0, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2,-1}, // Machine Guns only
      {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0, 0, 1, 0, 1, 1,-1,-1,-1,-1,-1,-1}, // Rifles only
      {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 2, 2, 0, 1,-1,-1}, // Snipers only
      {-1,-1,-1, 2, 2, 0, 1, 2, 2, 2, 1, 2, 0, 2, 0, 0, 1, 0, 1, 1, 2, 2, 0, 1, 2, 1}  // Standard
   };

   int tabMapAS[7][Const_NumWeapons] =
   {
      {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}, // Knife only
      {-1,-1,-1, 2, 2, 0, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}, // Pistols only
      {-1,-1,-1,-1,-1,-1,-1, 1, 1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1}, // Shotgun only
      {-1,-1,-1,-1,-1,-1,-1,-1,-1, 1, 1, 1, 0, 2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 1,-1}, // Machine Guns only
      {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0,-1, 1, 0, 1, 1,-1,-1,-1,-1,-1,-1}, // Rifles only
      {-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, 0, 0,-1, 1,-1,-1}, // Snipers only
      {-1,-1,-1, 2, 2, 0, 1, 1, 1, 1, 1, 1, 0, 2, 0,-1, 1, 0, 1, 1, 0, 0,-1, 1, 1, 1}  // Standard
   };

   char modeName[7][12] =
   {
      {"Knife"},
      {"Pistol"},
      {"Shotgun"},
      {"Machine Gun"},
      {"Rifle"},
      {"Sniper"},
      {"Standard"}
   };
   selection--;

   for (int i = 0; i < Const_NumWeapons; i++)
   {
      g_weaponSelect[i].teamStandard = tabMapStandart[selection][i];
      g_weaponSelect[i].teamAS = tabMapAS[selection][i];
   }

   if (selection == 0)
      g_botVar[Variable_KnifeMode]->SetInt (1);
   else
      g_botVar[Variable_KnifeMode]->SetInt (0);

   CenterPrint ("%s weapon mode selected", &modeName[selection][0]);
}

void BotManager::ListBots (void)
{
   // this function list's bots currently playing on the server

   ServerPrintNoTag ("%-3.5s %-9.13s %-17.18s %-3.4s %-3.4s %-3.4s", "index", "name", "personality", "team", "skill", "frags");

   for (int i = 0; i < GetMaxClients (); i++)
   {
      edict_t *player = INDEXENT (i);

      // is this player slot valid
      if (IsValidBot (player))
         ServerPrintNoTag ("[%-3.1d] %-9.13s %-17.18s %-3.4s %-3.1d %-3.1d", i, STRING (player->v.netname), GetBot (player)->m_personality == Personality_Rusher ? "rusher" : GetBot (player)->m_personality == Personality_Normal ? "normal" : "careful", GetTeam (player) != 0 ? "CT" : "T", GetBot (player)->m_skill, static_cast <int> (player->v.frags));
   }
}

int BotManager::GetBotsNum (void)
{
   // this function returns number of yapb's playing on the server

   int count = 0;

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL)
         count++;
   }
   return count;
}

int BotManager::GetHumansNum (void)
{
   // this function returns number of humans playing on the server

   int count = 0;

   for (int i = 0; i < GetMaxClients (); i++)
   {
      if ((g_clients[i].flags & ClientFlag_Used) && m_bots[i] == NULL)
         count++;
   }
   return count;
}


Bot *BotManager::GetHighestFragsBot (int team)
{
   Bot *highFragBot = NULL;

   int bestIndex = 0;
   float bestScore = -1;

   // search bots in this team
   for (int i = 0; i < GetMaxClients (); i++)
   {
      highFragBot = g_botManager->GetBot (i);

      if (highFragBot != NULL && IsAlive (highFragBot->GetEntity ()) && GetTeam (highFragBot->GetEntity ()) == team)
      {
         if (highFragBot->pev->frags > bestScore)
         {
            bestIndex = i;
            bestScore = highFragBot->pev->frags;
         }
      }
   }
   return GetBot (bestIndex);
}

void BotManager::CheckTeamEconomics (int team)
{
   // this function decides is players on specified team is able to buy primary weapons by calculating players
   // that have not enough money to buy primary (with economics), and if this result higher 80%, player is can't
   // buy primary weapons.

   if (!g_botVar[Variable_Economics]->GetBool ())
   {
      m_economicsGood[team] = true;
      return; // don't check economics while economics disable
   }

   int numPoorPlayers = 0;
   int numTeamPlayers = 0;

   // start calculating
   for (int i = 0; i < GetMaxClients (); i++)
   {
      if (m_bots[i] != NULL && GetTeam (m_bots[i]->GetEntity ()) == team)
      {
         if (m_bots[i]->m_moneyAmount <= g_botBuyEconomyTable[0])
            numPoorPlayers++;

         numTeamPlayers++; // update count of team
      }
   }
   m_economicsGood[team] = true;

   if (numTeamPlayers <= 1)
      return;

   // if 80 percent of team have no enough money to purchase primary weapon
   if ((numTeamPlayers * 80) / 100 <= numPoorPlayers)
      m_economicsGood[team] = false;

   // winner must buy something!
   if (m_lastWinner == team)
      m_economicsGood[team] = true;
}

void BotManager::Free (void)
{
   // this function free all bots slots (used on server shutdown)

   for (int i = 0; i < 32; i++)
   {
      if (m_bots[i] != NULL)
      {
         delete m_bots[i];
         m_bots[i] = NULL;
      }
   }
}

void BotManager::Free (int index)
{
   // this function frees one bot selected by index (used on bot disconnect)

   delete m_bots[index];
   m_bots[index] = NULL;
}

Bot::Bot (edict_t *bot, int skill, int personality, int team, int member)
{
   // this function does core operation of creating bot, it's called by CreateBot (),
   // when bot setup completed, (this is a bot class constructor)

   char rejectReason[128];
   int clientIndex = ENTINDEX (bot);

   memset (this, 0, sizeof (Bot));

   pev = VARS (bot);

   if (bot->pvPrivateData != NULL)
      FREE_PRIVATE (bot);

   bot->pvPrivateData = NULL;
   bot->v.frags = 0;

   // create the player entity by calling MOD's player function
   BotManager::CallGameEntity (&bot->v);

   // set all info buffer keys for this bot
   char *buffer = GET_INFOKEYBUFFER (bot);

   SET_CLIENT_KEYVALUE (clientIndex, buffer, "model", "");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "rate", "3500.000000");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "cl_updaterate", "20");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "cl_lw", "1");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "cl_lc", "1");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "tracker", "0");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "cl_dlmax", "128");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "friends", "0");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "dm", "0");
   SET_CLIENT_KEYVALUE (clientIndex, buffer, "_ah", "0");
   // SET_CLIENT_KEYVALUE (clientIndex, staticBuffer, "*bot", "1");

   if (g_gameVersion != Version_Old)
      SET_CLIENT_KEYVALUE (clientIndex, buffer, "_vgui_menus", "0");

   rejectReason[0] = 0; // reset the reject reason template string
   MDLL_ClientConnect (bot, "BOT", FormatBuffer ("127.0.0.%d", ENTINDEX (bot) + 100), rejectReason);

   if (!IsNullString (rejectReason))
   {
      AddLogEntry (true, Log_Warning, "Server refused '%s' connection (%s)", STRING (bot->v.netname), rejectReason);
      ServerCommand ("kick \"%s\"", STRING (bot->v.netname)); // kick the bot player if the server refused it

      bot->v.flags |= FL_KILLME;
   }

   if (IsDedicatedServer () && g_botVar[Variable_Developer]->GetBool ())
   {
      if (g_botVar[Variable_Developer]->GetInt () == 2)
      {
          ServerPrint ("Server requiring authentication");
          ServerPrint ("Client '%s' connected", STRING (bot->v.netname));
          ServerPrint ("Adr: 127.0.0.%d:27005", ENTINDEX (bot) + 100);
      }

      ServerPrint ("Verifying and uploading resources...");
      ServerPrint ("Custom resources total 0 bytes");
      ServerPrint ("  Decals:  0 bytes");
      ServerPrint ("----------------------");
      ServerPrint ("Resources to request: 0 bytes");
   }

   MDLL_ClientPutInServer (bot);
   bot->v.flags |= FL_FAKECLIENT; // set this player as fakeclient

   // initialize all the variables for this bot...
   m_notStarted = true;  // hasn't joined game yet

   m_startAction = GameMessage_Idle;
   m_moneyAmount = 0;
   m_logotypeIndex = g_randGen.Long (0, String (g_botVar[Variable_Logotypes]->GetString ()).Split (';').GetElementNumber () - 1);

   // initialize msec value
   m_msecNum = m_msecDel = 0.0;
   m_msecInterval = GetWorldTime ();
   m_msecVal = static_cast <byte> (g_pGlobals->frametime * 1000.0);

   // assign how talkative this bot will be
   m_sayTextBuffer.chatDelay = g_randGen.Float (3.8, 10.0);
   m_sayTextBuffer.chatProbability = g_randGen.Long (1, 100);

   m_notKilled = false;
   m_skill = skill;
   m_weaponBurstMode = BurstMode_Disabled;

   m_lastThinkTime = GetWorldTime ();
   m_frameInterval = GetWorldTime ();

   bot->v.idealpitch = bot->v.v_angle.x;
   bot->v.ideal_yaw = bot->v.v_angle.y;

   bot->v.yaw_speed = g_randGen.Float (g_skillTab[m_skill / 20].minTurnSpeed, g_skillTab[m_skill / 20].maxTurnSpeed);
   bot->v.pitch_speed = g_randGen.Float (g_skillTab[m_skill / 20].minTurnSpeed, g_skillTab[m_skill / 20].maxTurnSpeed);

   switch (personality)
   {
   case 1:
      m_personality = Personality_Rusher;
      m_baseAgressionLevel = g_randGen.Float (0.7, 1.0);
      m_baseFearLevel = g_randGen.Float (0.0, 0.4);
      break;

   case 2:
      m_personality = Personality_Careful;
      m_baseAgressionLevel = g_randGen.Float (0.0, 0.4);
      m_baseFearLevel = g_randGen.Float (0.7, 1.0);
      break;

   default:
      m_personality = Personality_Normal;
      m_baseAgressionLevel = g_randGen.Float (0.4, 0.7);
      m_baseFearLevel = g_randGen.Float (0.4, 0.7);
      break;
   }

   memset (&m_ammoInClip, 0, sizeof (m_ammoInClip));
   memset (&m_ammo, 0, sizeof (m_ammo));

   m_currentWeapon = 0; // current weapon is not assigned at start
   m_voicePitch = g_randGen.Long (166, 250) / 2; // assign voice pitch

   // copy them over to the temp level variables
   m_agressionLevel = m_baseAgressionLevel;
   m_fearLevel = m_baseFearLevel;
   m_nextEmotionUpdate = GetWorldTime () + 0.5;

   // just to be sure
   m_actMessageIndex = 0;
   m_pushMessageIndex = 0;

   // assign team and class
   m_wantedTeam = team;
   m_wantedClass = member;

   NewRound ();
}

Bot::~Bot (void)
{
   // this is bot destructor

   SwitchChatterIcon (false);
   DeleteSearchNodes ();
   ResetTasks ();

   // free used botname
   IterateArray (g_botNames, j)
   {
      if (strcmp (g_botNames[j].name, STRING (pev->netname)) == 0)
      {
         g_botNames[j].isUsed = false;
         break;
      }
   }
}

void Bot::NewRound (void)
{
   // this function initializes a bot after creation & at the start of each round

   int i = 0;

   // delete all allocated path nodes
   DeleteSearchNodes ();

   m_waypointOrigin = Vector::GetNull ();
   m_destOrigin = Vector::GetNull ();
   m_currentWaypointIndex = -1;
   m_currentTravelFlags = 0;
   m_desiredVelocity = Vector::GetNull ();
   m_prevGoalIndex = -1;
   m_chosenGoalIndex = -1;
   m_loosedBombWptIndex = -1;

   m_moveToC4 = false;
   m_duckDefuse = false;
   m_duckDefuseCheckTime = 0.0;

   m_prevWptIndex[0] = -1;
   m_prevWptIndex[1] = -1;
   m_prevWptIndex[2] = -1;
   m_prevWptIndex[3] = -1;
   m_prevWptIndex[4] = -1;

   m_navTimeset = GetWorldTime ();

   switch (m_personality)
   {
   case Personality_Normal:
      m_pathType = g_randGen.Long (0, 100) > 50 ? 0 : 1;
      break;

   case Personality_Rusher:
      m_pathType = 0;
      break;

   case Personality_Careful:
      m_pathType = 1;
      break;
   }

   // clear all states & tasks
   m_states = 0;
   ResetTasks ();

   m_isVIP = false;
   m_isLeader = false;
   m_hasProgressBar = false;
   m_canChooseAimDirection = true;

   m_timeTeamOrder = 0.0;
   m_askCheckTime = 0.0;
   m_minSpeed = 260.0;
   m_prevSpeed = 0.0;
   m_prevOrigin = Vector (9999.0, 9999.0, 9999.0);
   m_prevTime = GetWorldTime ();
   m_blindRecognizeTime = GetWorldTime ();
   
   m_viewDistance = 4096.0;
   m_maxViewDistance = 4096.0;

   m_liftEntity = NULL;
   m_pickupItem = NULL;
   m_itemIgnore = NULL;
   m_itemCheckTime = 0.0;

   m_breakableEntity = NULL;
   m_breakable = Vector::GetNull ();
   m_timeDoorOpen = 0.0;

   ResetCollideState ();
   ResetDoubleJumpState ();

   m_enemy = NULL;
   m_lastVictim = NULL;
   m_lastEnemy = NULL;
   m_lastEnemyOrigin = Vector::GetNull ();
   m_trackingEdict = NULL;
   m_timeNextTracking = 0.0;

   m_buttonPushTime = 0.0;
   m_enemyUpdateTime = 0.0;
   m_seeEnemyTime = 0.0;
   m_shootAtDeadTime = 0.0;
   m_oldCombatDesire = 0.0;
   m_liftUsageTime = 0.0;

   m_avoidGrenade = NULL;
   m_needAvoidGrenade = 0;

   m_lastDamageType = -1;
   m_voteKickIndex = 0;
   m_lastVoteKick = 0;
   m_voteMap = 0;
   m_doorOpenAttempt = 0;
   m_aimFlags = 0;
   m_liftState = 0;
   m_burstShotsFired = 0;

   m_position = Vector::GetNull ();
   m_liftTravelPos = Vector::GetNull ();

   m_idealReactionTime = g_skillTab[m_skill / 20].minSurpriseTime;
   m_actualReactionTime = g_skillTab[m_skill / 20].minSurpriseTime;

   m_targetEntity = NULL;
   m_followWaitTime = 0.0;

   for (i = 0; i < Const_MaxHostages; i++)
      m_hostages[i] = NULL;

   for (i = 0; i < Chatter_Total; i++)
      m_voiceTimers[i] = -1.0;

   m_isReloading = false;
   m_reloadState = ReloadState_None;

   m_reloadCheckTime = 0.0;
   m_shootTime = GetWorldTime ();
   m_playerTargetTime = GetWorldTime ();
   m_firePause = 0.0;
   m_timeLastFired = 0.0;

   m_grenadeCheckTime = 0.0;
   m_isUsingGrenade = false;

   m_skillOffset = static_cast <int> ((100 - m_skill) / 100.0);
   m_blindButton = 0;
   m_blindTime = 0.0;
   m_jumpTime = 0.0;
   m_jumpFinished = false;
   m_isStuck = false;

   m_sayTextBuffer.timeNextChat = GetWorldTime ();
   m_sayTextBuffer.entityIndex = -1;
   m_sayTextBuffer.sayText[0] = 0x0;

   m_buyState = 0;

   if (!m_notKilled) // if bot died, clear all weapon stuff and force buying again
   {
      memset (&m_ammoInClip, 0, sizeof (m_ammoInClip));
      memset (&m_ammo, 0, sizeof (m_ammo));

      m_currentWeapon = 0;
   }

   m_knifeAttackTime = GetWorldTime () + g_randGen.Float (1.3, 2.6);
   m_nextBuyTime = GetWorldTime () + g_randGen.Float (0.6, 1.2);

   m_buyPending = false;
   m_inBombZone = false;

   m_shieldCheckTime = 0.0;
   m_zoomCheckTime = 0.0;
   m_strafeSetTime = 0.0;
   m_combatStrafeDir = 0;
   m_fightStyle = 0;
   m_lastFightStyleCheck = 0.0;

   m_checkWeaponSwitch = true;
   m_checkKnifeSwitch = true;
   m_buyingFinished = false;

   m_radioEntity = NULL;
   m_radioOrder = 0;
   m_defendedBomb = false;

   m_timeLogoSpray = GetWorldTime () + g_randGen.Float (0.5, 2.0);
   m_spawnTime = GetWorldTime ();
   m_lastChatTime = GetWorldTime ();
   pev->v_angle.y = pev->ideal_yaw;

   m_timeCamping = 0;
   m_campDirection = 0;
   m_nextCampDirTime = 0;
   m_campButtons = 0;

   m_soundUpdateTime = 0.0;
   m_heardSoundTime = GetWorldTime ();

   // clear its message queue
   for (i = 0; i < 32; i++)
      m_messageQueue[i] = GameMessage_Idle;

   m_actMessageIndex = 0;
   m_pushMessageIndex = 0;

   // and put buying into its message queue
   PushMessageQueue (GameMessage_BuyStuff);
   StartTask (Task_Normal, TaskPriority_Normal, -1, 0.0, true);
}

void Bot::Kill (void)
{
   // this function kills a bot (not just using ClientKill, but like the CSBot does)
   // base code courtesy of Lazy (from bots-united forums!)

   edict_t *hurtEntity = (*g_engfuncs.pfnCreateNamedEntity) (MAKE_STRING ("trigger_hurt"));

   if (FNullEnt (hurtEntity))
      return;

   hurtEntity->v.classname = MAKE_STRING (g_weaponDefs[m_currentWeapon].className);
   hurtEntity->v.dmg_inflictor = GetEntity ();
   hurtEntity->v.dmg = 9999.0;
   hurtEntity->v.dmg_take = 1.0;
   hurtEntity->v.dmgtime = 2.0;
   hurtEntity->v.effects |= EF_NODRAW;

   (*g_engfuncs.pfnSetOrigin) (hurtEntity, Vector (-4000, -4000, -4000));

   KeyValueData kv;
   kv.szClassName = const_cast <char *> (g_weaponDefs[m_currentWeapon].className);
   kv.szKeyName = "damagetype";
   kv.szValue = FormatBuffer ("%d", (1 << 4));
   kv.fHandled = false;

   MDLL_KeyValue (hurtEntity, &kv);

   MDLL_Spawn (hurtEntity);
   MDLL_Touch (hurtEntity, GetEntity ());

   (*g_engfuncs.pfnRemoveEntity) (hurtEntity);
}

void Bot::Kick (void)
{
   // this function kick off one bot from the server.

   ServerCommand ("kick #%d", GETPLAYERUSERID (GetEntity ()));
   CenterPrint ("Bot '%s' kicked", STRING (pev->netname));

   // balances quota
   if (g_botManager->GetBotsNum () - 1 < g_botVar[Variable_Quota]->GetInt ())
      g_botVar[Variable_Quota]->SetInt (g_botManager->GetBotsNum () - 1);
}

void Bot::StartGame (void)
{
   // this function handles the selection of teams & class

   // handle counter-strike stuff here...
   if (m_startAction == GameMessage_TeamSelect)
   {
      m_startAction = GameMessage_Idle;  // switch back to idle

      if (g_botVar[Variable_ForceTeam]->GetString ()[0] == 'C' || g_botVar[Variable_ForceTeam]->GetString ()[0] == 'c')
         m_wantedTeam = 2;
      else if (g_botVar[Variable_ForceTeam]->GetString ()[0] == 'T' || g_botVar[Variable_ForceTeam]->GetString ()[0] == 't')
         m_wantedTeam = 1;

      if (m_wantedTeam != 1 && m_wantedTeam != 2)
         m_wantedTeam = 5;

      // select the team the bot wishes to join...
      FakeClientCommand (GetEntity (), "menuselect %d", m_wantedTeam);
   }
   else if (m_startAction == GameMessage_ClassSelect)
   {
      m_startAction = GameMessage_Idle;  // switch back to idle

      if (g_gameVersion == Version_ConZero) // czero has spetsnaz and militia skins
      {
         if (m_wantedClass < 1 || m_wantedClass > 5)
            m_wantedClass = g_randGen.Long (1, 5); //  use random if invalid
      }
      else
      {
         if (m_wantedClass < 1 || m_wantedClass > 4)
            m_wantedClass = g_randGen.Long (1, 4); // use random if invalid
      }

      // select the class the bot wishes to use...
      FakeClientCommand (GetEntity (), "menuselect %d", m_wantedClass);

      // bot has now joined the game (doesn't need to be started)
      m_notStarted = false;

      // check for greeting other players, since we connected
      if (g_randGen.Long (0, 100) < 20)
         ChatMessage (Chat_Welcome);
   }
}
