//
// Copyright (c) 2007-2008, by YaPB Development Team. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
// Thanks to:
//  Whistler - Original author of YaPB.
//  KWo - Current developer of POD-Bot MM.
//  Evilspy - Author of Metamod-p, and jk-botti.
//
// Version: $Id:$
//

#ifndef YAPB_INCLUDED
#define YAPB_INCLUDED

#include <extdll.h>
#include <stdio.h>
#include <memory.h>

#include <dllapi.h>
#include <meta_api.h>

using namespace Math;

// detects the build platform
#if defined (__linux__) || defined (__debian__) || defined (__linux)
   #define PLATFORM_LINUX32 1
#elif defined (__x86_64__) || defined (__amd64__)
   #define PLATFORM_LINUX64 1
#elif defined (_WIN32)
   #define PLATFORM_WIN32 1
#endif

// detects the compiler
#if defined (_MSC_VER)
   #define COMPILER_VISUALC _MSC_VER
#elif defined (__BORLANDC__)
   #define COMPILER_BORLAND __BORLANDC__
#elif defined (__MINGW32__)
   #define COMPILER_MINGW32 __MINGW32__
#endif

// configure export macros
#if defined (COMPILER_VISUALC) || defined (COMPILER_MINGW32)
   #define export extern "C" __declspec (dllexport)
#elif defined (PLATFORM_LINUX32) || defined (PLATFORM_LINUX64) || defined (COMPILER_BORLAND)
   #define export extern "C"
#else
   #error "Can't configure export macros. Compiler unrecognized."
#endif

// operating system specific macros, functions and typedefs
#ifdef PLATFORM_WIN32

   #include <direct.h>

   #define DLL_ENTRYPOINT int STDCALL DllMain (HINSTANCE hModule, DWORD dwReason, LPVOID)
   #define DLL_DETACHING (dwReason == DLL_PROCESS_DETACH)
   #define DLL_RETENTRY return TRUE

   #if defined (COMPILER_VISUALC)
      #define DLL_GIVEFNPTRSTODLL extern "C" void STDCALL
   #elif defined (COMPILER_MINGW32)
      #define DLL_GIVEFNPTRSTODLL export void STDCALL
   #endif

   // specify export parameter
   #if defined (COMPILER_VISUALC) && (COMPILER_VISUALC > 1000)
      #pragma comment (linker, "/EXPORT:GiveFnptrsToDll=_GiveFnptrsToDll@8,@1")
      #pragma comment (linker, "/SECTION:.data,RW")
   #endif

   typedef int (FAR *EntityAPI_t) (gamefuncs_t *, int);
   typedef int (FAR *NewEntityAPI_t) (newgamefuncs_t *, int *);
   typedef int (FAR *BlendAPI_t) (int, void **, void *, float (*)[3][4], float (*)[128][3][4]);
   typedef void (STDCALL *FuncPointers_t) (enginefuncs_t *, globalvars_t *);
   typedef void (FAR *EntityPtr_t) (entvars_t *);

#elif defined (PLATFORM_LINUX32) || defined (PLATFORM_LINUX64)

   #include <unistd.h>
   #include <dlfcn.h>
   #include <errno.h>
   #include <sys/stat.h>

   #define DLL_ENTRYPOINT void _fini (void)
   #define DLL_DETACHING TRUE
   #define DLL_RETENTRY return
   #define DLL_GIVEFNPTRSTODLL extern "C" void

   inline uint32 _lrotl (uint32 x, int r) { return (x << r) | (x >> (sizeof (x) * 8 - r));}

   typedef int (*EntityAPI_t) (gamefuncs_t *, int);
   typedef int (*NewEntityAPI_t) (newgamefuncs_t *, int *);
   typedef int (*BlendAPI_t) (int, void **, void *, float (*)[3][4], float (*)[128][3][4]);
   typedef void (*FuncPointers_t) (enginefuncs_t *, globalvars_t *);
   typedef void (*EntityPtr_t) (entvars_t *);

#else
   #error "Platform unrecognized."
#endif

// library wrapper
class Library
{
private:
   void *m_ptr;
   
public:

   Library (const char *fileName)
   {
      if (fileName == NULL)
         return;

#ifdef PLATFORM_WIN32
      m_ptr = LoadLibrary (fileName);
#else
      m_ptr = dlopen (fileName, RTLD_NOW);
#endif
   }

 ~Library (void)
  {
      if (!IsLoaded ())
         return;

#ifdef PLATFORM_WIN32
      FreeLibrary ((HMODULE) m_ptr);
#else
      dlclose (m_ptr);
#endif
  }

public:
   void *GetFunctionAddr (const char *functionName)
   {
      if (!IsLoaded ())
         return NULL;

#ifdef PLATFORM_WIN32
      return GetProcAddress ((HMODULE) m_ptr, functionName);
#else
      return dlsym (m_ptr, functionName);
#endif
   }

   inline bool IsLoaded (void) const
   {
      return m_ptr != NULL;
   }
};

#include <assert.h>
#include <ctype.h>
#include <limits.h>
#include <float.h>
#include <time.h>

#include <corelib.h>

// defines bots tasks
enum TaskId_t
{
   Task_Normal,
   Task_Pause,
   Task_MoveToPosition,
   Task_FollowUser,
   Task_WaitForGo,
   Task_PickupItem,
   Task_Camp,
   Task_PlantBomb,
   Task_DefuseBomb,
   Task_AttackEnemy,
   Task_HuntEnemy,
   Task_SeekCover,
   Task_ThrowExpGrenade,
   Task_ThrowFlashGrenade,
   Task_ThrowSmokeGrenade,
   Task_PerformDoubleJump,
   Task_EscapeFromBomb,
   Task_ShootBreakable,
   Task_Hide,
   Task_Blinded,
   Task_SprayLogo
};

// supported cs's
enum ModSupportIndex_t
{
   Version_Steam = 1, // Counter-Strike 1.6 and Above
   Version_ConZero = 2, // Counter-Strike: Condition Zero
   Version_Old = 3 // Counter-Strike 1.3-1.5 with/without Steam
};

// log levels
enum LogLevel_t
{
   Log_Default = 1, // default log message
   Log_Warning = 2, // warning log message
   Log_Error = 3, // error log message
   Log_Ignore = 4, // additional flag
   Log_Critical = 5  // fatal error log message (terminate the game!)

};

// chat types id's
enum ChatType_t
{
   Chat_Killing = 0, // id to kill chat array
   Chat_Dead, // id to dead chat array
   Chat_BombPlant, // id to bomb chat array
   Chat_TeamAttack, // id to team-attack chat array
   Chat_TeamKill, // id to team-kill chat array
   Chat_Welcome, // id to welcome chat array
   Chat_NoKeyword, // id to no keyword chat array
   Chat_Total // number for array
};

// personalities defines
enum Personality_t
{
   Personality_Normal = 0,
   Personality_Rusher,
   Personality_Careful
};

// collision states
enum CollisionState_t
{
   Collision_NotDecided,
   Collision_Probing,
   Collision_NoMove,
   Collision_Jump,
   Collision_Duck,
   Collision_StrafeLeft,
   Collision_StrafeRight
};

// counter-strike team id's
enum Team_t
{
   Team_Terrorist = 0,
   Team_Counter
};

// client flags
enum ClientFlags_t
{
   ClientFlag_Used  = (1 << 0),
   ClientFlag_Alive = (1 << 1),
   ClientFlag_Admin = (1 << 2)
};

// bot cvars
enum BotVariable_t
{
   Variable_AutoVacate = 0,
   Variable_Quota = 1,
   Variable_MinSkill = 2,
   Variable_MaxSkill = 3,
   Variable_FollowUser = 4,
   Variable_TimerSound = 5,
   Variable_TimerPickup = 6,
   Variable_TimerGrenade = 7,
   Variable_DebugGoal = 8,
   Variable_Chat = 9,
   Variable_Synth = 10,
   Variable_KnifeMode = 11,
   Variable_SkillTags = 12,
   Variable_Stop = 13,
   Variable_ThruWalls = 14,
   Variable_Votes = 15,
   Variable_Spray = 16,
   Variable_BotBuy = 17,
   Variable_Debug = 18,
   Variable_TurnMethod = 19,
   Variable_Prefix = 20,
   Variable_Language = 21,
   Variable_Version = 22,
   Variable_DamperCoeff_X = 23,
   Variable_DamperCoeff_Y = 24,
   Variable_Deviation_X = 25,
   Variable_Deviation_Y = 26,
   Variable_Influence_XonY = 27,
   Variable_Influence_YonX = 28,
   Variable_SlowdownRatio = 29,
   Variable_OffsetDelay = 30,
   Variable_SpringStiff_X = 31,
   Variable_SpringStiff_Y = 32,
   Variable_AnticipRatio = 33,
   Variable_Password = 34,
   Variable_PasswordKey = 35,
   Variable_ChatterPath = 36,
   Variable_TKPunish = 37,
   Variable_Communication = 38,
   Variable_Economics = 39,
   Variable_ForceTeam = 40,
   Variable_RestWeapons = 41,
   Variable_DangerFactor = 42,
   Variable_DontShoot = 43,
   Variable_HardcoreMode = 44,
   Variable_AutoFollow = 45,
   Variable_Logotypes = 46,
   Variable_MsecMethod = 47,
   Variable_WptSubFolder = 48,
   Variable_LogLevel = 49,
   Variable_WalkAllowed = 50,
   Variable_Deathmatch = 51,
   Variable_ThinkFPS = 52,
   Variable_C4Timer = 53,
   Variable_BuyTime = 54,
   Variable_FriendlyFire = 55,
   Variable_RoundTime = 56,
   Variable_Gravity = 57,
   Variable_Developer = 58,
   Variable_FreezeTime = 59,
   Variable_FootSteps = 60,
   Variable_Total
};

// radio messages
enum RadioMessage_t
{
   Radio_CoverMe = 1,
   Radio_YouTakePoint = 2,
   Radio_HoldPosition = 3,
   Radio_RegroupTeam = 4,
   Radio_FollowMe = 5,
   Radio_TakingFire = 6,
   Radio_GoGoGo = 11,
   Radio_Fallback = 12,
   Radio_StickTogether = 13,
   Radio_GetInPosition = 14,
   Radio_StormTheFront = 15,
   Radio_ReportTeam = 16,
   Radio_Affirmative = 21,
   Radio_EnemySpotted = 22,
   Radio_NeedBackup = 23,
   Radio_SectorClear = 24,
   Radio_InPosition = 25,
   Radio_ReportingIn = 26,
   Radio_ShesGonnaBlow = 27,
   Radio_Negative = 28,
   Radio_EnemyDown = 29
};

// voice system (extending enum above, messages 30-39 is reserved)
enum ChatterMessage_t
{
   Chatter_SpotTheBomber = 40,
   Chatter_FriendlyFire,
   Chatter_DiePain,
   Chatter_GotBlinded,
   Chatter_GoingToPlantBomb,
   Chatter_RescuingHostages,
   Chatter_GoingToCamp,
   Chatter_HearSomething,
   Chatter_TeamKill,
   Chatter_ReportingIn,
   Chatter_GuardDroppedC4,
   Chatter_Camp,
   Chatter_PlantingC4,
   Chatter_DefusingC4,
   Chatter_InCombat,
   Chatter_SeeksEnemy,
   Chatter_Nothing,
   Chatter_EnemyDown,
   Chatter_UseHostage,
   Chatter_FoundC4,
   Chatter_WonTheRound,
   Chatter_ScaredEmotion,
   Chatter_HeardEnemy,
   Chatter_SniperWarning,
   Chatter_SniperKilled,
   Chatter_VIPSpotted,
   Chatter_GuardingVipSafety,
   Chatter_GoingToGuardVIPSafety,
   Chatter_QuicklyWonTheRound,
   Chatter_OneEnemyLeft,
   Chatter_TwoEnemiesLeft,
   Chatter_ThreeEnemiesLeft,
   Chatter_NoEnemiesLeft,
   Chatter_FoundBombPlace,
   Chatter_WhereIsTheBomb,
   Chatter_DefendingBombSite,
   Chatter_BarelyDefused,
   Chatter_NiceshotCommander,
   Chatter_NiceshotPall,
   Chatter_GoingToGuardHostages,
   Chatter_GoingToGuardDoppedBomb,
   Chatter_OnMyWay,
   Chatter_LeadOnSir,
   Chatter_Total
};

// counter-strike weapon id's
enum Weapon_t
{
   Weapon_SigP228 = 1,
   Weapon_ShieldGun = 2,
   Weapon_Scout = 3,
   Weapon_Esplosive = 4,
   Weapon_Xm1014 = 5,
   Weapon_BombC4 = 6,
   Weapon_Mac10 = 7,
   Weapon_Aug = 8,
   Weapon_Smoke = 9,
   Weapon_Elite = 10,
   Weapon_FiveSeven = 11,
   Weapon_Ump45 = 12,
   Weapon_Sig550 = 13,
   Weapon_Galil = 14,
   Weapon_Famas = 15,
   Weapon_Usp = 16,
   Weapon_Glock18 = 17,
   Weapon_Magnum = 18,
   Weapon_Mp5Navy = 19,
   Weapon_M249 = 20,
   Weapon_M3 = 21,
   Weapon_M4A1 = 22,
   Weapon_Tmp = 23,
   Weapon_G3SG1 = 24,
   Weapon_Flashbang = 25,
   Weapon_Deagle = 26,
   Weapon_Sig552 = 27,
   Weapon_Ak47 = 28,
   Weapon_Knife = 29,
   Weapon_P90 = 30,
   Weapon_Armor = 31,
   Weapon_ArmorHelm = 32,
   Weapon_Defuser = 33
};

// defines for pickup items
enum PickupType_t
{
   PickupType_None,
   PickupType_Weapon,
   PickupType_DroppedC4,
   PickupType_PlantedC4,
   PickupType_Hostage,
   PickupType_Button,
   PickupType_ShieldGun,
   PickupType_DefusalKit
};

// reload state
enum ReloadState_t
{
   ReloadState_None = 0, // no reload state currrently
   ReloadState_Primary = 1, // primary weapon reload state
   ReloadState_Secondary = 2  // secondary weapon reload state
};

// collision probes
enum CollisionProbe_t
{
   Probe_Jump = (1 << 0), // probe jump when colliding
   Probe_Duck = (1 << 1), // probe duck when colliding
   Probe_Strafe = (1 << 2) // probe strafing when colliding
};

// vgui menus (since latest steam updates is obsolete, but left for old cs)
enum VGuiMenu_t
{
   MenuState_Team = 2, // menu select team
   MenuState_Terrorist = 26, // terrorist select menu
   MenuState_Counter = 27  // ct select menu
};

// lift usage states
enum LiftState_t
{
   LiftState_NoNearby = 0,
   LiftState_LookingButtonOutside,
   LiftState_WaitingFor,
   LiftState_GoingIn,
   LiftState_WaitingForTeammates,
   LiftState_LookingButtonInside,
   LiftState_TravelingBy,
   LiftState_Leaving
};

// game start messages for counter-strike...
enum GameMessage_t
{
   GameMessage_Idle = 1,
   GameMessage_TeamSelect = 2,
   GameMessage_ClassSelect = 3,
   GameMessage_BuyStuff = 100,
   GameMessage_RadioMsg = 200,
   GameMessage_SayMsg = 10000,
   GameMessage_TeamSay = 10001
};

// netmessage functions
enum NetworkMessage_t
{
   NetMessage_VGui = 1,
   NetMessage_ShowMenu = 2,
   NetMessage_WeaponList = 3,
   NetMessage_CurWeapon = 4,
   NetMessage_AmmoX = 5,
   NetMessage_AmmoPickup = 6,
   NetMessage_Damage = 7,
   NetMessage_Money = 8,
   NetMessage_StatusIcon = 9,
   NetMessage_Death = 10,
   NetMessage_ScreenFade = 11,
   NetMessage_HLTV = 12,
   NetMessage_TextMsg = 13,
   NetMessage_ScoreInfo = 14,
   NetMessage_BarTime = 15,
   NetMessage_SendAudio = 17,
   NetMessage_SayText = 18,
   NetMessage_BotVoice = 19,
   NetMessage_ResetHud = 20,
   NetMessage_Undefined = 0
};

// sensing states
enum SensingState_t
{
   State_SeeingEnemy = (1 << 0), // seeing an enemy
   State_HearingEnemy = (1 << 1), // hearing an enemy
   State_PickupItem = (1 << 2), // pickup item nearby
   State_ThrowExplosive = (1 << 3), // could throw he grenade
   State_ThrowFlashbang = (1 << 4), // could throw flashbang
   State_ThrowSmoke = (1 << 5), // could throw smokegrenade
   State_SuspectEnemy = (1 << 6) // suspect enemy behind obstacle
};

// positions to aim at
enum AimPosition_t
{
   AimPos_NavPoint = (1 << 0), // aim at nav point
   AimPos_CampPoint = (1 << 1), // aim at camp vector
   AimPos_PredictEnemy = (1 << 2), // aim at predicted path
   AimPos_LastEnemy = (1 << 3), // aim at last enemy
   AimPos_Entity = (1 << 4), // aim at entity like buttons, hostages
   AimPos_Enemy = (1 << 5), // aim at enemy
   AimPos_Grenade = (1 << 6), // aim for grenade throw
   AimPos_Override = (1 << 7)  // overrides all others (blinded)
};

// famas/glock burst mode status + m4a1/usp silencer
enum BurstMode_t
{
   BurstMode_Enabled  = 1,
   BurstMode_Disabled = 2
};

// visibility flags
enum Visibility_t
{
   Visibility_Head = (1 << 1),
   Visibility_Body = (1 << 2),
   Visibility_Other = (1 << 3)
};

// defines map type
enum MapType_t
{
   MapType_As = (1 << 0),
   MapType_Cs = (1 << 1),
   MapType_De = (1 << 2),
   MapType_Es = (1 << 3),
   MapType_Ka = (1 << 4),
   MapType_Fy = (1 << 5)
};

// defines for waypoint flags field (32 bits are available)
enum WaypointFlag_t
{
   PointFlag_Lift = (1 << 1), // wait for lift to be down before approaching this waypoint
   PointFlag_Crouch = (1 << 2), // must crouch to reach this waypoint
   PointFlag_Crossing = (1 << 3), // a target waypoint
   PointFlag_Goal = (1 << 4), // mission goal point (bomb, hostage etc.)
   PointFlag_Ladder = (1 << 5), // waypoint is on ladder
   PointFlag_Rescue = (1 << 6), // waypoint is a hostage rescue point
   PointFlag_Camp = (1 << 7), // waypoint is a camping point
   PointFlag_NoHostage = (1 << 8), // only use this waypoint if no hostage
   PointFlag_DoubleJump = (1 << 9), // bot help's another bot (requster) to get somewhere (using djump)
   PointFlag_Sniper = (1 << 28), // it's a specific sniper point
   PointFlag_Terrorist = (1 << 29), // it's a specific terrorist point
   PointFlag_Counter = (1 << 30)  // it's a specific ct point
};

// defines for waypoint connection flags field (16 bits are available)
enum PathFlag_t
{
   PathFlag_Jump = (1 << 0), // must jump for this connection
   PathFlag_DoubleJump = (1 << 1) // must use friend for this connection
};

// defines waypoint connection types
enum PathConnection_t
{
   PointConn_Outgoing = 0,
   PointConn_Incoming,
   PointConn_Bothways
};

// bot known file headers
const char FileHeader_Waypoint[] = "PODWAY!";
const char FileHeader_Experience[] = "PODEXP!";
const char FileHeader_Vistable[] = "PODVIS!";

const int FileVersion_Waypoint = 7;
const int FileVersion_Experience = 3;
const int FileVersion_Vistable = 1;

// some hardcoded desire defines used to override calculated ones
const float TaskPriority_Normal = 35.0;
const float TASKPRI_PAUSE = 36.0;
const float TASKPRI_CAMP  = 37.0;
const float TASKPRI_SPRAYLOGO = 38.0;
const float TASKPRI_FOLLOWUSER = 39.0;
const float TASKPRI_MOVETOPOSITION = 50.0;
const float TASKPRI_DEFUSEBOMB = 89.0;
const float TASKPRI_PLANTBOMB = 89.0;
const float TASKPRI_ATTACK = 90.0;
const float TASKPRI_SEEKCOVER = 91.0;
const float TASKPRI_HIDE = 92.0;
const float TASKPRI_THROWGRENADE = 99.0;
const float TASKPRI_DOUBLEJUMP = 99.0;
const float TASKPRI_BLINDED = 100.0;
const float TASKPRI_SHOOTBREAKABLE = 100.0;
const float TASKPRI_ESCAPEFROMBOMB = 100.0;

const int Const_GrenadeTimer = 3;
const int Const_MaxHostages = 8;
const int Const_MaxPathIndex = 8;
const int Const_MaxDamageValue = 2040;
const int Const_MaxGoalValue = 2040;
const int Const_MaxKillHistory = 16;
const int Const_MaxRegMessages = 256;
const int Const_MaxWaypoints = 1024;
const int Const_MaxWeapons = 32;
const int Const_NumWeapons = 26;

// weapon masks
const int WeaponBits_Primary = ((1 << Weapon_Xm1014) | (1 <<Weapon_M3) | (1 << Weapon_Mac10) | (1 << Weapon_Ump45) | (1 << Weapon_Mp5Navy) | (1 << Weapon_Tmp) | (1 << Weapon_P90) | (1 << Weapon_Aug) | (1 << Weapon_M4A1) | (1 << Weapon_Sig552) | (1 << Weapon_Ak47) | (1 << Weapon_Scout) | (1 << Weapon_Sig550) | (1 << Weapon_Magnum) | (1 << Weapon_G3SG1) | (1 << Weapon_M249) | (1 << Weapon_Famas) | (1 << Weapon_Galil));
const int WeaponBits_Secondary = ((1 << Weapon_SigP228) | (1 << Weapon_Elite) | (1 << Weapon_Usp) | (1 << Weapon_Glock18) | (1 << Weapon_Deagle) | (1 << Weapon_FiveSeven));

// this structure links waypoints returned from pathfinder
struct PathNode_t
{
   int index;
   PathNode_t *next;
};

// links keywords and replies together
struct KeywordChat_t
{
   Array <String> keywords;
   Array <String> replies;
   Array <String> usedReplies;
};

// tasks definition
struct Task_t
{
   Task_t *prev; // previuous task in the linked list
   Task_t *next; // next task in the linked list
   TaskId_t taskID; // major task/action carried out
   float desire; // desire (filled in) for this task
   int data; // additional data (waypoint index)
   float time; // time task expires
   bool canContinue; // if task can be continued if interrupted
};

// wave structure
struct WaveHeader_t
{
   char riffChunkId[4];
   unsigned long packageSize;
   char chunkID[4];
   char formatChunkId[4];
   unsigned long formatChunkLength;
   unsigned short dummy;
   unsigned short channels;
   unsigned long sampleRate;
   unsigned long bytesPerSecond;
   unsigned short bytesPerSample;
   unsigned short bitsPerSample;
   char dataChunkId[4];
   unsigned long dataChunkLength;
};

// botname structure definition
struct BotName_t
{
   String name;
   bool isUsed;
};

// voice config structure definition
struct Chatter_t
{
   String name;
   float repeatTime;
};

// language config structure definition
struct Language_t
{
   char *original; // original string
   char *translated; // string to replace for
};

struct WeaponSelect_t
{
   int id; // the weapon id value
   char *weaponName; // name of the weapon when selecting it
   char *modelName; // model name to separate cs weapons
   int price; // price when buying
   int minPrimaryAmmo; // minimum primary ammo
   int teamStandard; // used by team (number) (standard map)
   int teamAS; // used by team (as map)
   int buyGroup; // group in buy menu (standard map)
   int buySelect; // select item in buy menu (standard map)
   int newBuySelectT; // for counter-strike v1.6
   int newBuySelectCT; // for counter-strike v1.6
   bool shootsThru; // can shoot thru walls
   bool primaryFireHold; // hold down primary fire button to use?
};

// skill definitions
struct SkillDefinition_t
{
   float minSurpriseTime; // surprise delay (minimum)
   float maxSurpriseTime; // surprise delay (maximum)
   float campStartDelay; // delay befor start camping
   float campEndDelay; // delay before end camping
   float minTurnSpeed; // initial minimum turnspeed
   float maxTurnSpeed; // initial maximum turnspeed
   float aimOffs_X; // X/Y/Z maximum offsets
   float aimOffs_Y; // X/Y/Z maximum offsets
   float aimOffs_Z; // X/Y/Z maximum offsets
   int headshotFrequency; // precent to aiming to player head
   int heardShootThruProb; // precent to shooting throug wall when seen something
   int seenShootThruProb; // precent to shooting throug wall when heard something
   int recoilAmount; // amount of recoil when the bot should pause shooting
};

// fire delay definiton
struct FireDelay_t
{
   int weaponIndex;
   int maxFireBullets;
   float minBurstPauseFactor;
   float primaryBaseDelay;
   float primaryMinDelay[6];
   float primaryMaxDelay[6];
   float secondaryBaseDelay;
   float secondaryMinDelay[5];
   float secondaryMaxDelay[5];
};

// struct for menus
struct MenuText_t
{
   int validSlots; // ored together bits for valid keys
   char *menuText; // ptr to actual string
};

// array of clients struct
struct Client_t
{
   MenuText_t *menu; // pointer to opened bot menu
   edict_t *ent; // pointer to actual edict
   Vector origin; // position in the world
   Vector soundPosition; // position sound was played
   int team; // bot team
   int realTeam; // real bot team in free for all mode (csdm)
   int flags; // client flags
   float hearingDistance; // distance this sound is heared
   float timeSoundLasting; // time sound is played/heared
   float maxTimeSoundLasting; // max time sound is played/heared (to divide the difference between that above one and the current one)
};

// experience data hold in memory while playing
struct Experience_t
{
   unsigned short team0Damage;
   unsigned short team1Damage;
   signed short team0DangerIndex;
   signed short team1DangerIndex;
   signed short team0Value;
   signed short team1Value;
};

// experience data when saving/loading
struct ExperienceSave_t
{
   unsigned char team0Damage;
   unsigned char team1Damage;
   signed char team0Value;
   signed char team1Value;
};

// bot creation tab
struct CreateQueue_t
{
   int skill;
   int team;
   int member;
   int personality;
   String name;
};

// weapon properties structure
struct WeaponProperty_t
{
   char className[64];
   int ammo1; // ammo index for primary ammo
   int ammo1Max; // max primary ammo
   int slotID; // HUD slot (0 based)
   int position; // slot position
   int id; // weapon ID
   int flags; // flags???
};

// define chatting collection structure
struct SayText_t
{
   char chatProbability;
   float chatDelay;
   float timeNextChat;
   int entityIndex;
   char sayText[512];
   Array <String> lastUsedSentences;
};

// general waypoint header information structure
struct WaypointHeader_t
{
   char header[8];
   int32 fileVersion;
   int32 pointNumber;
   char mapName[32];
   char author[32];
};

// general experience & vistable header information structure
struct ExtensionHeader_t
{
   char header[8];
   int32 fileVersion;
   int32 pointNumber;
};

// define general waypoint structure
struct Path_t
{
   int32 pathNumber;
   int32 flags;
   Vector origin;
   float radius;

   float campStartX;
   float campStartY;
   float campEndX;
   float campEndY;

   int16 index[Const_MaxPathIndex];
   uint16 connectionFlags[Const_MaxPathIndex];
   Vector connectionVelocity[Const_MaxPathIndex];
   int32 distances[Const_MaxPathIndex];

   struct Vis_t { uint16 stand, crouch; } vis;
};

// main bot class
class Bot
{
   friend class BotManager;

private:
   unsigned int m_states; // sensing bitstates
   Task_t *m_tasks; // pointer to active tasks/schedules

   float m_moveSpeed; // current speed forward/backward
   float m_strafeSpeed; // current speed sideways
   float m_minSpeed; // minimum speed in normal mode
   float m_oldCombatDesire; // holds old desire for filtering

   bool m_isLeader; // bot is leader of his team
   bool m_checkTerrain; // check for terrain
   bool m_moveToC4; // ct is moving to bomb

   float m_prevTime; // time previously checked movement speed
   float m_prevSpeed; // speed some frames before
   Vector m_prevOrigin; // origin some frames before

   int m_messageQueue[32]; // stack for messages
   char m_tempStrings[160]; // space for strings (say text...)
   int m_radioSelect; // radio entry

   float m_blindRecognizeTime; // time to recognize enemy
   float m_itemCheckTime; // time next search for items needs to be done
   PickupType_t m_pickupType; // type of entity which needs to be used/picked up
   Vector m_breakable; // origin of breakable

   edict_t *m_pickupItem; // pointer to entity of item to use/pickup
   edict_t *m_itemIgnore; // pointer to entity to ignore for pickup
   edict_t *m_liftEntity; // pointer to lift entity
   edict_t *m_breakableEntity; // pointer to breakable entity

   float m_timeDoorOpen; // time to next door open check
   float m_lastChatTime; // time bot last chatted
   float m_timeLogoSpray; // time bot last spray logo
   float m_knifeAttackTime; // time to rush with knife (at the beginning of the round)
   bool m_defendedBomb; // defend action issued

   float m_askCheckTime; // time to ask team
   float m_collideTime; // time last collision
   float m_firstCollideTime; // time of first collision
   float m_probeTime; // time of probing different moves
   float m_lastCollTime; // time until next collision check
   char m_collisionProbeBits; // bits of possible collision moves
   char m_collideMoves[4]; // sorted array of movements
   char m_collStateIndex; // index into collide moves
   CollisionState_t m_collisionState; // collision State

   PathNode_t *m_navNode; // pointer to current node from path
   PathNode_t *m_navNodeStart; // pointer to start of path finding nodes
   unsigned char m_pathType; // which pathfinder to use
   unsigned char m_visibility; // visibility flags

   int m_currentWaypointIndex; // current waypoint index
   int m_travelStartIndex; // travel start index to double jump action
   int m_prevWptIndex[5]; // previous waypoint indices from waypoint find
   int m_waypointFlags; // current waypoint flags
   int m_loosedBombWptIndex; // nearest to loosed bomb waypoint
   int m_skillOffset; // offset to bots skill

   unsigned short m_currentTravelFlags; // connection flags like jumping
   bool m_jumpFinished; // has bot finished jumping
   Vector m_desiredVelocity; // desired velocity for jump waypoints
   float m_navTimeset; // time waypoint chosen by Bot

   unsigned int m_aimFlags; // aiming conditions
   Vector m_lookAt; // vector bot should look at
   Vector m_throw; // origin of waypoint to throw grenades

   Vector m_enemyOrigin; // target origin chosen for shooting
   Vector m_grenade; // calculated vector for grenades
   Vector m_entity; // origin of entities like buttons etc.
   Vector m_camp; // aiming vector when camping.

   float m_timeWaypointMove; // last time bot followed waypoints
   bool m_wantsToFire; // bot needs consider firing
   float m_shootAtDeadTime; // time to shoot at dying players
   edict_t *m_avoidGrenade; // pointer to grenade entity to avoid
   char m_needAvoidGrenade; // which direction to strafe away

   float m_followWaitTime; // wait to follow time
   edict_t *m_targetEntity; // the entity that the bot is trying to reach
   edict_t *m_hostages[Const_MaxHostages]; // pointer to used hostage entities

   bool m_isStuck; // bot is stuck
   bool m_isReloading; // bot is reloading a gun
   int m_reloadState; // current reload state
   int m_voicePitch; // bot voice pitch

   bool m_duckDefuse; // should or not bot duck to defuse bomb
   float m_duckDefuseCheckTime; // time to check for ducking for defuse

   byte m_msecVal; // calculated msec value
   float m_msecDel; // used for msec calculation
   float m_msecNum; // also used for mseccalculation
   float m_msecInterval; // used for leon hartwig's method for msec calculation

   float m_frameInterval; // bot's frame interval
   float m_lastThinkTime; // time bot last thinked

   float m_reloadCheckTime; // time to check reloading
   float m_zoomCheckTime; // time to check zoom again
   float m_shieldCheckTime; // time to check shiled drawing again
   float m_grenadeCheckTime; // time to check grenade usage

   bool m_checkKnifeSwitch; // is time to check switch to knife action
   bool m_checkWeaponSwitch; // is time to check weapon switch
   bool m_isUsingGrenade; // bot currently using grenade??

   unsigned char m_combatStrafeDir; // direction to strafe
   unsigned char m_fightStyle; // combat style to use
   float m_lastFightStyleCheck; // time checked style
   float m_strafeSetTime; // time strafe direction was set

   float m_timeCamping; // time to camp
   int m_campDirection; // camp Facing direction
   float m_nextCampDirTime; // time next camp direction change
   int m_campButtons; // buttons to press while camping
   int m_doorOpenAttempt; // attempt's to open the door
   int m_liftState; // state of lift handling

   float m_duckTime; // time to duck
   float m_jumpTime; // time last jump happened
   float m_voiceTimers[Chatter_Total]; // voice command timers
   float m_soundUpdateTime; // time to update the sound
   float m_heardSoundTime; // last time noise is heard
   float m_buttonPushTime; // time to push the button
   float m_liftUsageTime; // time to use lift

   Vector m_liftTravelPos; // lift travel position
   Vector m_moveAngles; // bot move angles
   bool m_moveToGoal; // bot currently moving to goal??

   Vector m_idealAngles; // angle wanted
   Vector m_randomizedIdealAngles; // angle wanted with noise
   Vector m_angularDeviation; // angular deviation from current to ideal angles
   Vector m_aimSpeed; // aim speed calculated
   Vector m_targetOriginAngularSpeed; // target/enemy angular speed

   float m_randomizeAnglesTime; // time last randomized location
   float m_playerTargetTime; // time last targeting

   void SwitchChatterIcon (bool show);
   void InstantChatterMessage (int type);
   void BotAI (void);
   bool IsMorePowerfulWeaponCanBeBought (void);
   void PerformWeaponPurchase (void);

   bool CanDuckUnder (Vector normal);
   bool CanJumpUp (Vector normal);
   bool CanStrafeLeft (TraceResult *tr);
   bool CanStrafeRight (TraceResult *tr);
   bool CantMoveForward (Vector normal, TraceResult *tr);

   void ChangePitch (float speed);
   void ChangeYaw (float speed);
   void CheckMessageQueue (void);
   void CheckRadioCommands (void);
   void CheckReload (void);
   void AvoidGrenades (void);
   void CheckBurstMode (float distance);

   void CheckSilencer (void);
   bool CheckWallOnLeft (void);
   bool CheckWallOnRight (void);
   void ChooseAimDirection (void);
   int ChooseBombWaypoint (void);

   bool DoWaypointNav (void);
   bool EnemyIsThreat (void);
   void FacePosition (void);
   bool IsRestricted (int weaponIndex);
   bool IsRestrictedAMX (int weaponIndex);

   bool IsInViewCone (Vector origin);
   void ReactOnSound (void);
   bool CheckVisibility (entvars_t *targetOrigin, Vector *origin, byte *bodyPart);
   bool IsEnemyViewable (edict_t *player);

   edict_t *FindNearestButton (const char *className);
   edict_t *FindBreakable (void);
   int FindCoverWaypoint (float maxDistance);
   int FindDefendWaypoint (Vector origin);
   int FindGoal (void);
   void FindItem (void);

   void GetCampDirection (Vector *dest);
   void CollectGoalExperience (int damage, int team);
   void CollectExperienceData (edict_t *attacker, int damage);
   int GetMessageQueue (void);
   bool GoalIsValid (void);
   bool HeadTowardWaypoint (void);
   int InFieldOfView (Vector dest);

   bool IsBombDefusing (Vector bombOrigin);
   bool IsBlockedLeft (void);
   bool IsBlockedRight (void);
   bool IsWaypointUsed (int index);
   inline bool IsOnLadder (void) { return pev->movetype == MOVETYPE_FLY; }
   inline bool IsOnFloor (void) { return (pev->flags & (FL_ONGROUND | FL_PARTIALGROUND)) != 0; }
   inline bool IsInWater (void) { return pev->waterlevel >= 2; }
   inline float GetWalkSpeed (void) { return static_cast <float> ((static_cast <int> (pev->maxspeed) / 2) + (static_cast <int> (pev->maxspeed) / 50)) - 18; }
   
   bool ItemIsVisible (Vector dest, char *itemName);
   bool LastEnemyShootable (void);
   bool IsLastEnemyViewable (void);
   bool IsBehindSmokeClouds (edict_t *ent);
   void RunTask (void);
   void CheckTasksPriorities (void);
   void StartTask (Task_t *task);

   bool IsShootableBreakable (edict_t *ent);
   bool RateGroundWeapon (edict_t *ent);
   bool ReactOnEnemy (void);
   void ResetCollideState (void);
   void SetConditions (void);
   void SetStrafeSpeed (Vector moveDir, float strafeSpeed);
   void StartGame (void);
   void TaskComplete (void);
   bool GetBestNextWaypoint (void);
   int GetBestWeaponCarried (void);
   int GetBestSecondaryWeaponCarried (void);

   void RunPlayerMovement (void);
   void GetValidWaypoint (void);
   void ChangeWptIndex (int waypointIndex);
   bool IsDeadlyDrop (Vector targetOriginPos);
   bool OutOfBombTimer (void);
   void SelectLeaderEachTeam (int team);

   Vector CheckToss (const Vector &start, Vector end);
   Vector CheckThrow (const Vector &start, Vector end);
   Vector GetAimPosition (void);
   Vector CheckBombAudible (void);

   int CheckGrenades (void);
   void CommandTeam (void);
   void AttachToUser (void);
   void CombatFight (void);
   bool IsWeaponBadInDistance (int weaponIndex, float distance);
   bool DoFirePause (float distance, FireDelay_t *fireDelay);
   bool LookupEnemy (void);
   void FireWeapon (void);
   void FocusEnemy (void);

   void SelectBestWeapon (void);
   void SelectPistol (void);
   bool IsFriendInLineOfFire (float distance);
   bool IsGroupOfEnemies (Vector location, int numEnemies = 1, int radius = 256);
   bool IsShootableThruObstacle (Vector dest);
   int GetNearbyEnemiesNearPosition (Vector origin, int radius);
   int GetNearbyFriendsNearPosition (Vector origin, int radius);
   void SelectWeaponByName (const char *name);
   void SelectWeaponbyNumber (int num);
   int GetHighestWeapon (void);

   bool IsEnemyProtectedByShield (edict_t *enemy);
   bool ParseChat (char *reply);
   bool RepliesToPlayer (void);
   float GetBombTimeleft (void);
   float GetEstimatedReachTime (void);

   int GetAimingWaypoint (void);
   int GetAimingWaypoint (Vector targetOriginPos);
   void FindShortestPath (int srcIndex, int destIndex);
   void FindPath (int srcIndex, int destIndex, unsigned char pathType = 0);
   void DebugMsg (const char *format, ...);
   void SecondThink (void);

public:
   entvars_t *pev;

   int m_wantedTeam; // player team bot wants select
   int m_wantedClass; // player model bot wants to select

   int m_skill; ;// bots play skill
   int m_moneyAmount; // amount of money in bot's bank

   Personality_t m_personality;
   float m_spawnTime; // time this bot spawned
   float m_timeTeamOrder; // time of last radio command

   bool m_isVIP; // bot is vip?
   bool m_bIsDefendingTeam; // bot in defending team on this map

   int m_startAction; // team/class selection state
   bool m_notKilled; // has the player been killed or has he just respawned
   bool m_notStarted; // team/class not chosen yet

   int m_voteKickIndex; // index of player to vote against
   int m_lastVoteKick; // last index
   int m_voteMap; // number of map to vote for
   int m_logotypeIndex; // index for logotype
   int m_burstShotsFired; // number of bullets fired

   bool m_inBombZone; // bot in the bomb zone or not
   int m_buyState; // current Count in Buying
   float m_nextBuyTime; // next buy time

   bool m_inBuyZone; // bot currently in buy zone
   bool m_inVIPZone; // bot in the vip satefy zone
   bool m_buyingFinished; // done with buying
   bool m_buyPending; // bot buy is pending
   bool m_hasDefuser; // does bot has defuser
   bool m_hasProgressBar; // has progress bar on a HUD
   bool m_jumpReady; // is double jump ready
   bool m_canChooseAimDirection; // can choose aiming direction

   float m_blindTime; // time when bot is blinded
   float m_blindMoveSpeed; // mad speeds when bot is blind
   float m_blindSidemoveSpeed; // mad side move speeds when bot is blind
   int m_blindButton; // buttons bot press, when blind

   edict_t *m_doubleJumpEntity; // pointer to entity that request double jump
   edict_t *m_radioEntity; // pointer to entity issuing a radio command
   int m_radioOrder; // actual command

   float m_duckForJump; // is bot needed to duck for double jump
   float m_baseAgressionLevel; // base aggression level (on initializing)
   float m_baseFearLevel; // base fear level (on initializing)
   float m_agressionLevel; // dynamic aggression level (in game)
   float m_fearLevel; // dynamic fear level (in game)
   float m_nextEmotionUpdate; // next time to sanitize emotions

   int m_actMessageIndex; // current processed message
   int m_pushMessageIndex; // offset for next pushed message

   int m_prevGoalIndex; // holds destination goal waypoint
   int m_chosenGoalIndex; // used for experience, same as above
   float m_goalValue; // ranking value for this waypoint

   Vector m_waypointOrigin; // origin of waypoint
   Vector m_destOrigin; // origin of move destination
   Vector m_position; // position to move to in move to position task
   Vector m_doubleJumpOrigin; // origin of double jump
   Vector m_lastBombPosition; // origin of last remembered bomb position

   float m_viewDistance; // current view distance
   float m_maxViewDistance; // maximum view distance
   Vector m_lastEnemyOrigin; // vector to last enemy origin
   SayText_t m_sayTextBuffer; // holds the index & the actual message of the last unprocessed text message of a player
   BurstMode_t m_weaponBurstMode; // bot using burst mode? (famas/glock18, but also silencer mode)

   edict_t *m_enemy; // pointer to enemy entity
   float m_enemyUpdateTime; // time to check for new enemies
   float m_enemyReachableTimer; // time to recheck if Enemy reachable
   bool m_isEnemyReachable; // direct line to enemy

   float m_seeEnemyTime; // time bot sees enemy
   float m_enemySurpriseTime; // time of surprise
   float m_idealReactionTime; // time of base reaction
   float m_actualReactionTime; // time of current reaction time

   edict_t *m_lastEnemy; // pointer to last enemy entity
   edict_t *m_lastVictim; // pointer to killed entity
   edict_t *m_trackingEdict; // pointer to last tracked player when camping/hiding
   float m_timeNextTracking; // time waypoint index for tracking player is recalculated

   float m_thinkTimer; // think timer interval
   float m_firePause; // time to pause firing
   float m_shootTime; // time to shoot
   float m_timeLastFired; // time to last firing
   int m_lastDamageType; // stores last damage

   int m_currentWeapon; // one current weapon for each bot
   int m_ammoInClip[Const_MaxWeapons]; // ammo in clip for each weapons
   int m_ammo[MAX_AMMO_SLOTS]; // total ammo amounts

   Bot (edict_t *bot, int skill, int personality, int team, int member);
  ~Bot (void);

   int GetAmmo (void);
   inline int GetAmmoInClip (void) { return m_ammoInClip[m_currentWeapon]; }

   inline edict_t *GetEntity (void) { return ENT (pev); };
   inline EOFFSET GetOffset (void) { return OFFSET (pev); };
   inline int GetIndex (void) { return ENTINDEX (GetEntity ()); };

   inline Vector Center (void) { return (pev->absmax + pev->absmin) * 0.5; };
   inline Vector EyePosition (void) { return pev->origin + pev->view_ofs; };
   inline Vector EarPosition (void) { return pev->origin + pev->view_ofs; };
   inline Vector GetGunPosition (void) { return (pev->origin + pev->view_ofs); }

   void Think (void);
   void NewRound (void);
   void EquipInBuyzone (int buyCount);
   void PushMessageQueue (int message);
   void PrepareChatMessage (char *text);
   bool FindWaypoint (void);
   bool EntityIsVisible (Vector dest, bool fromBody = false);

   void DeleteSearchNodes (void);
   Task_t *GetCurrentTask (void);

   void RemoveCertainTask (TaskId_t taskID);
   void ResetTasks (void);
   void TakeDamage (edict_t *inflictor, int damage, int armor, int bits);
   void TakeBlinded (Vector fade, int alpha);
   void StartTask (TaskId_t taskID, float desire, int data, float time, bool canContinue);
   void DiscardWeaponForUser (edict_t *user, bool discardC4);

   void SayText (const char *text);
   void TeamSayText (const char *text);

   void ChatMessage (int type, bool isTeamSay = false);
   void RadioMessage (int message);
   void ChatterMessage (int message);
   void HandleChatterMessage (const char *sz);

   void Kill (void);
   void Kick (void);
   void ResetDoubleJumpState (void);
   void MoveToVector (Vector to);
   int FindLoosedBomb (void);

   bool HasHostage (void);
   bool UsesRifle (void);
   bool UsesPistol (void);
   bool UsesSniper (void);
   bool UsesSubmachineGun (void);
   bool UsesZoomableRifle (void);
   bool UsesBadPrimary (void);
   bool HasPrimaryWeapon (void);
   bool HasShield (void);
   bool IsShieldDrawn (void);
};

// manager class
class BotManager : public Singleton <BotManager>
{
private:
   Array <CreateQueue_t> m_creationTab; // bot creation tab
   Bot *m_bots[32]; // all available bots
   float m_maintainTime; // time to maintain bot creation quota
   int m_lastWinner; // the team who won previous round
   int m_roundCount; // rounds passed
   bool m_economicsGood[2]; // is team able to buy anything

protected:
   int CreateBot (String name, int skill, int personality, int team, int member);

public:
   BotManager (void);
  ~BotManager (void);

   bool EconomicsValid (int team) { return m_economicsGood[team]; }

   int GetLastWinner (void) const { return m_lastWinner; }
   void SetLastWinner (int winner) { m_lastWinner = winner; }

   int GetIndex (edict_t *ent);
   Bot *GetBot (int index);
   Bot *GetBot (edict_t *ent);
   Bot *FindOneValidAliveBot (void);
   Bot *GetHighestFragsBot (int team);

   int GetHumansNum (void);
   int GetBotsNum (void);

   void Think (void);
   void Free (void);
   void Free (int index);
   void CheckAutoVacate (edict_t *ent);

   void AddRandom (void) { AddBot ("", -1, -1, -1, -1); }
   void AddBot (String name, int skill, int personality, int team, int member);
   void AddBot (String name, String skill, String personality, String team, String member);
   void FillServer (int selection, int personality = Personality_Normal, int skill = -1, int numToAdd = -1);

   void RemoveAll (void);
   void RemoveRandom (void);
   void RemoveFromTeam (Team_t team, bool removeAll = false);
   void RemoveMenu (edict_t *ent, int selection);
   void KillAll (int team = -1);
   void MaintainBotQuota (void);
   void InitQuota (void);

   void ListBots (void);
   void SetWeaponMode (int selection);
   void CheckTeamEconomics (int team);

   static void CallGameEntity (entvars_t *vars);
};

// texts localizer
class Localizer : public Singleton <Localizer>
{
public:
   Array <Language_t> m_langTab;

public:
   Localizer (void) { m_langTab.RemoveAll (); }
  ~Localizer (void) { m_langTab.RemoveAll (); }

   char *TranslateInput (const char *input);
};

// netmessage handler class
class NetworkMsg : public Singleton <NetworkMsg>
{
private:
   Bot *m_bot;
   int m_state;
   int m_message;
   int m_registerdMessages[NetMessage_BotVoice + 1];

public:
   NetworkMsg (void);
  ~NetworkMsg   (void) { };

   void Execute (void *p);
   void Reset (void) { m_message = NetMessage_Undefined; m_state = 0; m_bot = NULL; };
   void HandleMessageIfRequired (int messageType, int requiredType);

   void SetMessage (int message) { m_message = message; }
   void SetBot (Bot *bot) { m_bot = bot; }

   int GetId (int messageType) { return m_registerdMessages[messageType]; }
   void SetId (int messageType, int messsageIdentifier) { m_registerdMessages[messageType] = messsageIdentifier; }
};

// waypoint operation class
class Waypoint : public Singleton <Waypoint>
{
   friend class Bot;

private:
   Path_t *m_paths[Const_MaxWaypoints];

   bool m_waypointPaths;
   bool m_isOnLadder;
   bool m_endJumpPoint;
   bool m_learnJumpWaypoint;
   float m_timeJumpStarted;

   Vector m_learnVelocity;
   Vector m_learnPosition;
   Vector m_foundBombOrigin;

   int m_cacheWaypointIndex;
   int m_lastJumpWaypoint;
   int m_visibilityIndex;
   Vector m_lastWaypoint;
   byte m_visLUT[Const_MaxWaypoints][Const_MaxWaypoints / 4];

   float m_pathDisplayTime;
   float m_arrowDisplayTime;
   float m_waypointDisplayTime[Const_MaxWaypoints];
   float m_goalsScore[Const_MaxWaypoints];
   int m_findWPIndex;
   int m_facingAtIndex;
   char m_infoBuffer[256];

   int *m_distMatrix;
   int *m_pathMatrix;

   Array <int> m_terrorPoints;
   Array <int> m_ctPoints;
   Array <int> m_goalPoints;
   Array <int> m_campPoints;
   Array <int> m_sniperPoints;
   Array <int> m_rescuePoints;
   Array <int> m_visitedGoals;

public:
   bool m_redoneVisibility;

   Waypoint (void);
  ~Waypoint (void);

   void Init (void);
   void InitExperienceTab (void);
   void InitVisibilityTab (void);

   void InitTypes (void);
   void AddPath (short int addIndex,  short int pathIndex, float distance);

   int GetFacingIndex (void);
   int FindFarest (Vector origin, float maxDistance = 32.0);
   int FindNearest (Vector origin, float minDistance = 9999.0, int flags = -1);
   void FindInRadius (Vector origin, float radius, int *holdTab, int *count);
   void FindInRadius (Array <int> &queueID, float radius, Vector origin);

   void Add (int flags, Vector waypointOrigin = Vector::GetNull ());
   void Delete (void);
   void ToggleFlags (int toggleFlag);
   void SetRadius (int radius);
   bool IsConnected (int pointA, int pointB);
   bool IsConnected (int num);
   void InitializeVisibility (void);
   void CreatePath (char dir);
   void DeletePath (void);
   void CacheWaypoint (void);

   float GetTravelTime (float maxSpeed, Vector src, Vector origin);
   bool IsVisible (int srcIndex, int destIndex);
   bool IsStandVisible (int srcIndex, int destIndex);
   bool IsDuckVisible (int srcIndex, int destIndex);
   void CalculateWayzone (int index);

   bool Load (void);
   void Save (void);
   void SaveXML (void);

   bool Reachable (Bot *bot, int index);
   bool IsNodeReachable (Vector src, Vector destination);
   void Think (void);
   bool NodesValid (void);
   void SaveExperienceTab (void);
   void SaveVisibilityTab (void);
   void CreateBasic (void);
   void EraseFromHardDisk (void);

   void InitPathMatrix (void);
   void SavePathMatrix (void);
   bool LoadPathMatrix (void);

   int GetPathDistance (int srcIndex, int destIndex);
   Path_t *GetPath (int id);
   char *GetWaypointInfo (int id);
   char *GetInfo (void) { return m_infoBuffer; }

   int AddGoalScore (int index, int other[4]);
   void SetFindIndex (int index);
   void SetLearnJumpWaypoint (void);
   void ClearGoalScore (void);

   bool IsGoalVisited (int index);
   void SetGoalVisited (int index);

   Vector GetBombPosition (void) { return m_foundBombOrigin; }
   void SetBombPosition (bool shouldReset = false);
   String CheckSubfolderFile (void);
};

#define g_netMsg NetworkMsg::GetObject ()
#define g_botManager BotManager::GetObject ()
#define g_localizer Localizer::GetObject ()
#define g_waypoint Waypoint::GetObject ()

// prototypes of bot functions...
extern int GetMaxClients (void);
extern int GetWeaponReturn (bool isString, const char *weaponAlias, int weaponIndex = -1);
extern int GetTeam (edict_t *ent);

extern float GetShootingConeDeviation (edict_t *ent, Vector *position);
extern float GetWorldTime (void);
extern float GetWaveLength (const char *fileName);

extern bool IsLinux (void);
extern bool TryFileOpen (char *fileName);
extern bool IsDedicatedServer (void);
extern bool IsVisible (const Vector &origin, edict_t *ent);
extern bool IsAlive (edict_t *ent);
extern bool IsInViewCone (Vector origin, edict_t *ent);
extern bool IsWeaponShootingThroughWall (int id);
extern bool IsValidBot (edict_t *ent);
extern bool IsValidPlayer (edict_t *ent);
extern bool OpenConfig (const char *fileName, char *errorIfNotExists, File *outFile, bool languageDependant = false);
extern bool FindNearestPlayer (void **holder, edict_t *to, float searchDistance = 4096.0, bool sameTeam = false, bool needBot = false, bool needAlive = false, bool needDrawn = false);

extern const char *GetMapName (void);
extern const char *GetWaypointDir (void);
extern const char *GetModName (void);
extern const char *GetField (const char *string, int fieldId, bool endLine = false);
extern char *FormatBuffer (char *format, ...);

extern uint16 GenerateBuildNumber (void);
extern Vector GetEntityOrigin (edict_t *ent);

extern void FreeLibraryMemory (void);
extern void RoundInit (void);
extern void FakeClientCommand (edict_t *fakeClient, const char *format, ...);
extern void strtrim (char *string);
extern void CreatePath (char *path);
extern void ServerCommand (const char *format, ...);
extern void RegisterBotVariable (int variableID, const char *initialValue, int flags = FCVAR_EXTDLL | FCVAR_SERVER);
extern void RegisterCommand (char *command, void funcPtr (void));
extern void CheckWelcomeMessage (void);
extern void DetectCSVersion (void);
extern void PlaySound (edict_t *ent, const char *soundName);
extern void ServerPrint (const char *format, ...);
extern void ChartPrint (const char *format, ...);
extern void ServerPrintNoTag (const char *format, ...);
extern void CenterPrint (const char *format, ...);
extern void ClientPrint (edict_t *ent, int dest, const char *format, ...);
extern void HudMessage (edict_t *ent, bool toCenter, Vector rgb, char *format, ...);

extern void AddLogEntry (bool outputToConsole, int logLevel, const char *format, ...);
extern void TraceLine (const Vector &start, const Vector &end, bool ignoreMonsters, bool ignoreGlass, edict_t *ignoreEntity, TraceResult *ptr);
extern void TraceLine (const Vector &start, const Vector &end, bool ignoreMonsters, edict_t *ignoreEntity, TraceResult *ptr);
extern void TraceHull (const Vector &start, const Vector &end, bool ignoreMonsters, int hullNumber, edict_t *ignoreEntity, TraceResult *ptr);
extern void DrawLine (edict_t *ent, Vector start, Vector end, int width, int noise, int red, int green, int blue, int brightness, int speed, int life);
extern void DrawArrow (edict_t *ent, Vector start, Vector end, int width, int noise, int red, int green, int blue, int brightness, int speed, int life);
extern void DisplayMenuToClient (edict_t *ent, MenuText_t *menu);
extern void DecalTrace (entvars_t *pev, TraceResult *trace, int logotypeIndex);
extern void SoundAttachToThreat (edict_t *ent, const char *sample, float volume);
extern void SoundSimulateUpdate (int playerIndex);

inline bool IsNullString (const char *input)
{
   if (input == NULL)
      return true;

   return *input == '\0';
}

template <typename T> inline T *AllocateMembuf (int actualSize = sizeof (T))
{
   if (actualSize == sizeof (T))
      return new T;

   return new T[actualSize];
}

#include <globals.h>
#include <compress.h>
#include <resource.h>

#endif // YAPB_INCLUDED
